@@ -19,15 +19,19 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include <stdlib.h>
 #include <math.h>
 #include <string.h>
 
+#include "pango-impl-utils.h"
 #include "pango-types.h"
 #include "pango-font.h"
 #include "pango-fontmap.h"
 #include "pango-utils.h"
 
+static const char bad_font_warning[] = "%s called with bad font, expect ugly output";
+
 struct _PangoFontDescription
 {
   char *family_name;
@@ -39,7 +43,8 @@ struct _PangoFontDescription
 
   guint16 mask;
   guint static_family : 1;
-
+  guint size_is_absolute : 1;
+	
   int size;
 };
 
@@ -49,33 +54,43 @@ pango_font_description_get_type (void)
   static GType our_type = 0;
   
   if (our_type == 0)
-    our_type = g_boxed_type_register_static ("PangoFontDescription",
+    our_type = g_boxed_type_register_static (I_("PangoFontDescription"),
 					     (GBoxedCopyFunc)pango_font_description_copy,
 					     (GBoxedFreeFunc)pango_font_description_free);
 
   return our_type;
 }
 
+
+static const PangoFontDescription pfd_defaults = {
+  NULL,			/* family_name */
+
+  PANGO_STYLE_NORMAL,	/* style */
+  PANGO_VARIANT_NORMAL,	/* variant */
+  PANGO_WEIGHT_NORMAL,	/* weight */
+  PANGO_STRETCH_NORMAL,	/* stretch */
+
+  0,			/* mask */
+  0,			/* static_family */
+  FALSE,		/* size_is_absolute */
+
+  0,			/* size */
+};
+
 /**
  * pango_font_description_new:
  * 
  * Creates a new font description structure with all fields unset.
  * 
- * Return value: the newly-created #PangoFontDescription. Use
- * pango_font_description_free() to free the result.
+ * Return value: the newly allocated #PangoFontDescription, which
+ *               should be freed using pango_font_description_free().
  **/
 PangoFontDescription *
 pango_font_description_new (void)
 {
-  PangoFontDescription *desc = g_new (PangoFontDescription, 1);
+  PangoFontDescription *desc = g_slice_new (PangoFontDescription);
 
-  desc->mask = 0;
-  desc->family_name = NULL;
-  desc->style = PANGO_STYLE_NORMAL;
-  desc->variant = PANGO_VARIANT_NORMAL;
-  desc->weight = PANGO_WEIGHT_NORMAL;
-  desc->stretch = PANGO_STRETCH_NORMAL;
-  desc->size = 0;
+  *desc = pfd_defaults;
 
   return desc;
 }
@@ -97,23 +112,9 @@ pango_font_description_set_family (PangoFontDescription *desc,
 {
   g_return_if_fail (desc != NULL);
 
-  if (desc->family_name == family)
-    return;
-  
-  if (desc->family_name && !desc->static_family)
-    g_free (desc->family_name);
-
+  pango_font_description_set_family_static (desc, family ? g_strdup (family) : NULL);
   if (family)
-    {
-      desc->family_name = g_strdup (family);
-      desc->mask |= PANGO_FONT_MASK_FAMILY;
-      desc->static_family = FALSE;
-    }
-  else
-    {
-      desc->family_name = NULL;
-      desc->mask &= ~PANGO_FONT_MASK_FAMILY;
-    }
+    desc->static_family = FALSE;
 }
 
 /**
@@ -143,12 +144,13 @@ pango_font_description_set_family_static (PangoFontDescription *desc,
   if (family)
     {
       desc->family_name = (char *)family;
-      desc->mask |= PANGO_FONT_MASK_FAMILY;
       desc->static_family = TRUE;
+      desc->mask |= PANGO_FONT_MASK_FAMILY;
     }
   else
     {
-      desc->family_name = NULL;
+      desc->family_name = pfd_defaults.family_name;
+      desc->static_family = pfd_defaults.static_family;
       desc->mask &= ~PANGO_FONT_MASK_FAMILY;
     }
 }
@@ -160,7 +162,9 @@ pango_font_description_set_family_static (PangoFontDescription *desc,
  * Gets the family name field of a font description. See
  * pango_font_description_set_family().
  * 
- * Return value: The family name field. (Will be %NULL if not previously set.)
+ * Return value: the family name field for the font description, or
+ *               %NULL if not previously set.  This has the same life-time
+ *               as the font description itself and should not be freed.
  **/
 G_CONST_RETURN char *
 pango_font_description_get_family (const PangoFontDescription *desc)
@@ -208,7 +212,7 @@ pango_font_description_set_style (PangoFontDescription *desc,
 PangoStyle
 pango_font_description_get_style (const PangoFontDescription *desc)
 {
-  g_return_val_if_fail (desc != NULL, PANGO_STYLE_NORMAL);
+  g_return_val_if_fail (desc != NULL, pfd_defaults.style);
 
   return desc->style;
 }
@@ -245,7 +249,7 @@ pango_font_description_set_variant (PangoFontDescription *desc,
 PangoVariant
 pango_font_description_get_variant (const PangoFontDescription *desc)
 {
-  g_return_val_if_fail (desc != NULL, PANGO_VARIANT_NORMAL);
+  g_return_val_if_fail (desc != NULL, pfd_defaults.variant);
   
   return desc->variant;
 }
@@ -284,7 +288,7 @@ pango_font_description_set_weight (PangoFontDescription *desc,
 PangoWeight
 pango_font_description_get_weight (const PangoFontDescription *desc)
 {
-  g_return_val_if_fail (desc != NULL, PANGO_WEIGHT_NORMAL);
+  g_return_val_if_fail (desc != NULL, pfd_defaults.weight);
   
   return desc->weight;
 }
@@ -321,7 +325,7 @@ pango_font_description_set_stretch (PangoFontDescription *desc,
 PangoStretch
 pango_font_description_get_stretch (const PangoFontDescription *desc)
 {
-  g_return_val_if_fail (desc != NULL, PANGO_STRETCH_NORMAL);
+  g_return_val_if_fail (desc != NULL, pfd_defaults.stretch);
 
   return desc->stretch;
 }
@@ -329,10 +333,16 @@ pango_font_description_get_stretch (const PangoFontDescription *desc)
 /**
  * pango_font_description_set_size:
  * @desc: a #PangoFontDescription
- * @size: the size for the font description in pango units.
- *        There are #PANGO_SCALE Pango units in one device unit (device unit is a point, for font sizes).
+ * @size: the size of the font in points, scaled by PANGO_SCALE. (That is,
+ *        a @size value of 10 * PANGO_SCALE is a 10 point font. The conversion
+ *        factor between points and device units depends on system configuration
+ *        and the output device. For screen display, a logical DPI of 96 is
+ *        common, in which case a 10 point font corresponds to a 10 * (96 / 72) = 13.3
+ *        pixel font. Use pango_font_description_set_absolute_size() if you need
+ *        a particular size in device units.
  * 
- * Sets the size field of a font description.
+ * Sets the size field of a font description in fractional points. This is mutually
+ * exclusive with pango_font_description_set_absolute_size(). 
  **/
 void
 pango_font_description_set_size (PangoFontDescription *desc,
@@ -341,6 +351,7 @@ pango_font_description_set_size (PangoFontDescription *desc,
   g_return_if_fail (desc != NULL);
   
   desc->size = size;
+  desc->size_is_absolute = FALSE;
   desc->mask |= PANGO_FONT_MASK_SIZE;
 }
 
@@ -351,20 +362,64 @@ pango_font_description_set_size (PangoFontDescription *desc,
  * Gets the size field of a font description.
  * See pango_font_description_get_size().
  * 
- * Return value: the size field for the font description in pango
- *   units.  (There are #PANGO_SCALE pango units in one device unit &mdash;
- *   for fonts, font points are the device unit.) Returns 0 if the stretch field has not
+ * Return value: the size field for the font description in points or device units.
+ *   You must call pango_font_description_get_size_is_absolute()
+ *   to find out which is the case. Returns 0 if the size field has not
  *   previously been set.  pango_font_description_get_set_fields() to
  *   find out if the field was explicitely set or not.
  **/
 gint
 pango_font_description_get_size (const PangoFontDescription *desc)
 {
-  g_return_val_if_fail (desc != NULL, 0);
+  g_return_val_if_fail (desc != NULL, pfd_defaults.size);
 
   return desc->size;
 }
 
+/**
+ * pango_font_description_set_absolute_size:
+ * @desc: a #PangoFontDescription
+ * @size: the new size, in Pango units. There are #PANGO_SCALE Pango units in one
+ *   device unit. For an output backend where a device unit is a pixel, a @size
+ *   value of 10 * PANGO_SCALE gives a 10 pixel font.
+ * 
+ * Sets the size field of a font description, in device units. This is mutually
+ * exclusive with pango_font_description_set_size().
+ *
+ * Since: 1.8
+ **/
+void
+pango_font_description_set_absolute_size (PangoFontDescription *desc,
+					  double                size)
+{
+  g_return_if_fail (desc != NULL);
+  
+  desc->size = size;
+  desc->size_is_absolute = TRUE;
+  desc->mask |= PANGO_FONT_MASK_SIZE;
+}
+
+/**
+ * pango_font_description_get_size_is_absolute:
+ * @desc: a #PangoFontDescription
+ * 
+ * Determines whether the size of the font is in points or device units.
+ * See pango_font_description_set_size() and pango_font_description_set_absolute_size().
+ * 
+ * Return value: whether the size for the font description is in
+ *   points or device units.  Use pango_font_description_get_set_fields() to
+ *   find out if the size field of the font description was explicitely set or not.
+ *
+ * Since: 1.8
+ **/
+gboolean
+pango_font_description_get_size_is_absolute (const PangoFontDescription *desc)
+{
+  g_return_val_if_fail (desc != NULL, pfd_defaults.size_is_absolute);
+
+  return desc->size_is_absolute;
+}
+
 /**
  * pango_font_description_get_set_fields:
  * @desc: a #PangoFontDescription
@@ -377,7 +432,7 @@ pango_font_description_get_size (const PangoFontDescription *desc)
 PangoFontMask
 pango_font_description_get_set_fields (const PangoFontDescription *desc)
 {
-  g_return_val_if_fail (desc != NULL, 0);
+  g_return_val_if_fail (desc != NULL, pfd_defaults.mask);
 
   return desc->mask;
 }
@@ -387,19 +442,22 @@ pango_font_description_get_set_fields (const PangoFontDescription *desc)
  * @desc: a #PangoFontDescription
  * @to_unset: bitmask of fields in the @desc to unset.
  * 
- * Unsets some of the fields in a #PangoFontDescription. Note that
- * this merely marks the fields cleared, it does not clear the
- * settings for those fields, to clear a family name set with
- * pango_font_description_set_family_static() so that it won't
- * be returned by subsequent calls to pango_font_description_get_family(),
- * you must actually call pango_font_description_set_family (desc, NULL);
+ * Unsets some of the fields in a #PangoFontDescription.  The unset
+ * fields will get back to their default values.
  **/
 void
 pango_font_description_unset_fields (PangoFontDescription *desc,
 				     PangoFontMask         to_unset)
 {
+  PangoFontDescription unset_desc;
+  
   g_return_if_fail (desc != NULL);
 
+  unset_desc = pfd_defaults;
+  unset_desc.mask = to_unset;
+
+  pango_font_description_merge_static (desc, &unset_desc, TRUE);
+
   desc->mask &= ~to_unset;
 }
 
@@ -421,29 +479,15 @@ pango_font_description_merge (PangoFontDescription       *desc,
 			      const PangoFontDescription *desc_to_merge,
 			      gboolean                    replace_existing)
 {
-  PangoFontMask new_mask;
-  
-  g_return_if_fail (desc != NULL);
+  gboolean family_merged = desc_to_merge->family_name && (replace_existing || !desc->family_name);
 
-  if (replace_existing)
-    new_mask = desc_to_merge->mask;
-  else
-    new_mask = desc_to_merge->mask & ~desc->mask;
+  pango_font_description_merge_static (desc, desc_to_merge, replace_existing);
 
-  if (new_mask & PANGO_FONT_MASK_FAMILY)
-    pango_font_description_set_family (desc, desc_to_merge->family_name);
-  if (new_mask & PANGO_FONT_MASK_STYLE)
-    desc->style = desc_to_merge->style;
-  if (new_mask & PANGO_FONT_MASK_VARIANT)
-    desc->variant = desc_to_merge->variant;
-  if (new_mask & PANGO_FONT_MASK_WEIGHT)
-    desc->weight = desc_to_merge->weight;
-  if (new_mask & PANGO_FONT_MASK_STRETCH)
-    desc->stretch = desc_to_merge->stretch;
-  if (new_mask & PANGO_FONT_MASK_SIZE)
-    desc->size = desc_to_merge->size;
-
-  desc->mask |= new_mask;
+  if (family_merged)
+    {
+      desc->family_name = g_strdup (desc->family_name);
+      desc->static_family = FALSE;
+    }
 }
 
 /**
@@ -484,7 +528,10 @@ pango_font_description_merge_static (PangoFontDescription       *desc,
   if (new_mask & PANGO_FONT_MASK_STRETCH)
     desc->stretch = desc_to_merge->stretch;
   if (new_mask & PANGO_FONT_MASK_SIZE)
-    desc->size = desc_to_merge->size;
+    {
+      desc->size = desc_to_merge->size;
+      desc->size_is_absolute = desc_to_merge->size_is_absolute;
+    }
 
   desc->mask |= new_mask;
 }
@@ -510,7 +557,7 @@ compute_distance (const PangoFontDescription *a,
 
 /**
  * pango_font_description_better_match:
- * @desc: a #PangoFontDecription
+ * @desc: a #PangoFontDescription
  * @old_match: a #PangoFontDescription, or %NULL
  * @new_match: a #PangoFontDescription
  * 
@@ -548,18 +595,21 @@ pango_font_description_better_match (const PangoFontDescription *desc,
  * 
  * Make a copy of a #PangoFontDescription.
  * 
- * Return value: a newly-allocated #PangoFontDescription. This value
- *               must be freed using pango_font_description_free().
+ * Return value: the newly allocated #PangoFontDescription, which should
+ *               be freed with pango_font_description_free().
  **/
 PangoFontDescription *
 pango_font_description_copy  (const PangoFontDescription  *desc)
 {
-  PangoFontDescription *result = g_new (PangoFontDescription, 1);
+  PangoFontDescription *result = g_slice_new (PangoFontDescription);
 
   *result = *desc;
 
-  result->family_name = g_strdup (result->family_name);
-  result->static_family = FALSE;
+  if (result->family_name)
+    {
+      result->family_name = g_strdup (result->family_name);
+      result->static_family = FALSE;
+    }
 
   return result;
 }
@@ -573,16 +623,17 @@ pango_font_description_copy  (const PangoFontDescription  *desc)
  * be used until @desc is modififed or freed. This is meant to be used
  * when the copy is only needed temporarily.
  * 
- * Return value: a newly-allocated #PangoFontDescription. This value
- *               must be freed using pango_font_description_free().
+ * Return value: the newly allocated #PangoFontDescription, which should
+ *               be freed with pango_font_description_free().
  **/
 PangoFontDescription *
 pango_font_description_copy_static (const PangoFontDescription *desc)
 {
-  PangoFontDescription *result = g_new (PangoFontDescription, 1);
+  PangoFontDescription *result = g_slice_new (PangoFontDescription);
 
   *result = *desc;
-  result->static_family = TRUE;
+  if (result->family_name)
+    result->static_family = TRUE;
 
   return result;
 }
@@ -592,12 +643,14 @@ pango_font_description_copy_static (const PangoFontDescription *desc)
  * @desc1: a #PangoFontDescription
  * @desc2: another #PangoFontDescription
  * 
- * Compares two font descriptions for equality.
+ * Compares two font descriptions for equality. Two font descriptions
+ * are considered equal if the fonts they describe are provably identical.
+ * This means that their maskd do not have to match, as long as other fields
+ * are all the same. (Two font descrptions may result in identical fonts
+ * being loaded, but still compare %FALSE.)
  * 
- * Return value: %TRUE if the two font descriptions are proveably
- *               identical. (Two font descriptions may result in
- *               identical fonts being loaded, but still compare
- *                %FALSE.)
+ * Return value: %TRUE if the two font descriptions are identical,
+ * 		 %FALSE otherwise.
  **/
 gboolean
 pango_font_description_equal (const PangoFontDescription  *desc1,
@@ -606,14 +659,32 @@ pango_font_description_equal (const PangoFontDescription  *desc1,
   g_return_val_if_fail (desc1 != NULL, FALSE);
   g_return_val_if_fail (desc2 != NULL, FALSE);
 
-  return (desc1->mask == desc2->mask &&
-	  desc1->style == desc2->style &&
-	  desc1->variant == desc2->variant &&
-	  desc1->weight == desc2->weight &&
-	  desc1->stretch == desc2->stretch &&
-	  desc1->size == desc2->size &&
-	  (desc1->family_name == desc2->family_name ||
-	   (desc1->family_name && desc2->family_name && g_ascii_strcasecmp (desc1->family_name, desc2->family_name) == 0)));
+  return desc1->style == desc2->style &&
+	 desc1->variant == desc2->variant &&
+	 desc1->weight == desc2->weight &&
+	 desc1->stretch == desc2->stretch &&
+	 desc1->size == desc2->size &&
+	 desc1->size_is_absolute == desc2->size_is_absolute &&
+	 (desc1->family_name == desc2->family_name ||
+	  (desc1->family_name && desc2->family_name && g_ascii_strcasecmp (desc1->family_name, desc2->family_name) == 0));
+}
+
+#define TOLOWER(c) \
+  (((c) >= 'A' && (c) <= 'Z') ? (c) - 'A' + 'a' : (c))
+
+static guint
+case_insensitive_hash (const char *key)
+{
+  const char *p = key;
+  guint h = TOLOWER (*p);
+
+  if (h)
+    {
+      for (p += 1; *p != '\0'; p++)
+	h = (h << 5) - h + TOLOWER (*p);
+    }
+
+  return h;
 }
 
 /**
@@ -622,6 +693,7 @@ pango_font_description_equal (const PangoFontDescription  *desc1,
  * 
  * Computes a hash of a #PangoFontDescription structure suitable
  * to be used, for example, as an argument to g_hash_table_new().
+ * The hash value is independent of @desc->mask.
  * 
  * Return value: the hash value.
  **/
@@ -630,20 +702,14 @@ pango_font_description_hash (const PangoFontDescription *desc)
 {
   guint hash = 0;
 
-  hash = desc->mask;
-  
-  if (desc->mask & PANGO_FONT_MASK_FAMILY)
-    hash = g_str_hash (desc->family_name);
-  if (desc->mask & PANGO_FONT_MASK_SIZE)
-    hash ^= desc->size;
-  if (desc->mask & PANGO_FONT_MASK_STYLE)
-    hash ^= desc->style << 16;
-  if (desc->mask & PANGO_FONT_MASK_VARIANT)
-    hash ^= desc->variant << 18;
-  if (desc->mask & PANGO_FONT_MASK_WEIGHT)
-    hash ^= desc->weight << 16;
-  if (desc->mask & PANGO_FONT_MASK_STRETCH)
-    hash ^= desc->stretch << 26;
+  if (desc->family_name)
+    hash = case_insensitive_hash (desc->family_name);
+  hash ^= desc->size;
+  hash ^= desc->size_is_absolute ? 0xc33ca55a : 0;
+  hash ^= desc->style << 16;
+  hash ^= desc->variant << 18;
+  hash ^= desc->weight << 16;
+  hash ^= desc->stretch << 26;
 
   return hash;
 }
@@ -652,8 +718,7 @@ pango_font_description_hash (const PangoFontDescription *desc)
  * pango_font_description_free:
  * @desc: a #PangoFontDescription
  * 
- * Frees a font description returned from pango_font_describe()
- * or pango_font_description_copy().
+ * Frees a font description.
  **/
 void 
 pango_font_description_free  (PangoFontDescription  *desc)
@@ -663,7 +728,7 @@ pango_font_description_free  (PangoFontDescription  *desc)
       if (desc->family_name && !desc->static_family)
 	g_free (desc->family_name);
 
-      g_free (desc);
+      g_slice_free (PangoFontDescription, desc);
     }
 }
 
@@ -694,29 +759,29 @@ typedef struct
   const char *str;
 } FieldMap;
 
-static FieldMap style_map[] = {
+static const FieldMap style_map[] = {
   { PANGO_STYLE_NORMAL, NULL },
   { PANGO_STYLE_OBLIQUE, "Oblique" },
   { PANGO_STYLE_ITALIC, "Italic" }
 };
 
-static FieldMap variant_map[] = {
+static const FieldMap variant_map[] = {
   { PANGO_VARIANT_NORMAL, NULL },
   { PANGO_VARIANT_SMALL_CAPS, "Small-Caps" }
 };
 
-static FieldMap weight_map[] = {
+static const FieldMap weight_map[] = {
   { PANGO_WEIGHT_ULTRALIGHT, "Ultra-Light" },
   { PANGO_WEIGHT_LIGHT, "Light" },
   { PANGO_WEIGHT_NORMAL, NULL },
   { 500, "Medium" },
-  { 600, "Semi-Bold" },
+  { PANGO_WEIGHT_SEMIBOLD, "Semi-Bold" },
   { PANGO_WEIGHT_BOLD, "Bold" },
   { PANGO_WEIGHT_ULTRABOLD, "Ultra-Bold" },
   { PANGO_WEIGHT_HEAVY, "Heavy" }
 };
 
-static FieldMap stretch_map[] = {
+static const FieldMap stretch_map[] = {
   { PANGO_STRETCH_ULTRA_CONDENSED, "Ultra-Condensed" },
   { PANGO_STRETCH_EXTRA_CONDENSED, "Extra-Condensed" },
   { PANGO_STRETCH_CONDENSED,       "Condensed" },
@@ -729,13 +794,14 @@ static FieldMap stretch_map[] = {
 };
 
 static gboolean
-find_field (FieldMap *map, int n_elements, const char *str, int len, int *val)
+find_field (const FieldMap *map, int n_elements, const char *str, int len, int *val)
 {
   int i;
   
   for (i=0; i<n_elements; i++)
     {
-      if (map[i].str && g_ascii_strncasecmp (map[i].str, str, len) == 0)
+      if (map[i].str && map[i].str[0] && g_ascii_strncasecmp (map[i].str, str, len) == 0 &&
+	  map[i].str[len] == '\0')
 	{
 	  if (val)
 	    *val = map[i].value;
@@ -777,6 +843,25 @@ getword (const char *str, const char *last, size_t *wordlen)
   return result;
 }
 
+static gboolean
+parse_size (const char *word,
+	    size_t      wordlen,
+	    int        *pango_size)
+{
+  char *end;
+  double size = g_ascii_strtod (word, &end);
+
+  if ((size_t)(end - word) == wordlen) /* word is a valid float */
+    {
+      if (pango_size)
+	*pango_size = (int)(size * PANGO_SCALE + 0.5);
+
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
 /**
  * pango_font_description_from_string:
  * @str: string representation of a font description.
@@ -793,8 +878,7 @@ getword (const char *str, const char *last, size_t *wordlen)
  * default values. If SIZE is missing, the size in the resulting font
  * description will be set to 0.
  * 
- * Return value: a new #PangoFontDescription, or %NULL if the 
- * string could not be parsed. 
+ * Return value: a new #PangoFontDescription. 
  **/
 PangoFontDescription *
 pango_font_description_from_string (const char *str)
@@ -805,7 +889,7 @@ pango_font_description_from_string (const char *str)
 
   g_return_val_if_fail (str != NULL, NULL);
 
-  desc = g_new (PangoFontDescription, 1);
+  desc = g_slice_new (PangoFontDescription);
 
   desc->mask = PANGO_FONT_MASK_STYLE |
                PANGO_FONT_MASK_WEIGHT |
@@ -821,6 +905,7 @@ pango_font_description_from_string (const char *str)
   desc->stretch = PANGO_STRETCH_NORMAL;
   
   desc->size = 0;
+  desc->size_is_absolute = FALSE;
 
   len = strlen (str);
   last = str + len;
@@ -830,11 +915,8 @@ pango_font_description_from_string (const char *str)
    */
   if (wordlen != 0)
     {
-      char *end;
-      double size = g_ascii_strtod (p, &end);
-      if (end - p == wordlen) /* word is a valid float */
+      if (parse_size (p, wordlen, &desc->size))
 	{
-	  desc->size = (int)(size * PANGO_SCALE + 0.5);
 	  desc->mask |= PANGO_FONT_MASK_SIZE;
 	  last = p;
 	}
@@ -866,7 +948,7 @@ pango_font_description_from_string (const char *str)
   while (last > str && g_ascii_isspace (*(last - 1)))
     last--;
 
-  while (g_ascii_isspace (*str))
+  while (last > str && g_ascii_isspace (*str))
     str++;
 
   if (str != last)
@@ -879,7 +961,7 @@ pango_font_description_from_string (const char *str)
 }
 
 static void
-append_field (GString *str, FieldMap *map, int n_elements, int val)
+append_field (GString *str, const FieldMap *map, int n_elements, int val)
 {
   int i;
   for (i=0; i<n_elements; i++)
@@ -924,9 +1006,20 @@ pango_font_description_to_string (const PangoFontDescription  *desc)
       size_t wordlen;
 
       g_string_append (result, desc->family_name);
-      
+
+      /* We need to add a trailing comma if the family name ends
+       * in a keyword like "Bold", or if the family name ends in
+       * a number and no keywords will be added.
+       */
       p = getword (desc->family_name, desc->family_name + strlen(desc->family_name), &wordlen);
-      if (wordlen != 0 && find_field_any (p, wordlen, NULL))
+      if (wordlen != 0 &&
+	  (find_field_any (p, wordlen, NULL) ||
+	   (parse_size (p, wordlen, NULL) &&
+	    desc->weight == PANGO_WEIGHT_NORMAL &&
+	    desc->style == PANGO_STYLE_NORMAL &&
+	    desc->stretch == PANGO_STRETCH_NORMAL &&
+	    desc->variant == PANGO_VARIANT_NORMAL &&
+	    (desc->mask & PANGO_FONT_MASK_SIZE) == 0)))
 	g_string_append_c (result, ',');
     }
 
@@ -981,33 +1074,17 @@ pango_font_description_to_filename (const PangoFontDescription  *desc)
 
   return result;
 }  
-  
-GType
-pango_font_get_type (void)
-{
-  static GType object_type = 0;
 
-  if (!object_type)
-    {
-      static const GTypeInfo object_info =
-      {
-        sizeof (PangoFontClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        NULL,           /* class_init */
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoFont),
-        0,              /* n_preallocs */
-	NULL            /* init */
-      };
-      
-      object_type = g_type_register_static (G_TYPE_OBJECT,
-                                            "PangoFont",
-                                            &object_info, 0);
-    }
-  
-  return object_type;
+G_DEFINE_TYPE (PangoFont, pango_font, G_TYPE_OBJECT)
+
+static void
+pango_font_class_init (PangoFontClass *class)
+{
+}
+
+static void
+pango_font_init (PangoFont *font)
+{
 }
 
 /**
@@ -1048,7 +1125,7 @@ pango_font_get_coverage (PangoFont     *font,
  * pango_font_find_shaper:
  * @font: a #PangoFont
  * @language: the language tag
- * @ch: the ISO-10646 character code.
+ * @ch: a Unicode character.
  * 
  * Finds the best matching shaper for a font for a particular
  * language tag and character point.
@@ -1082,7 +1159,7 @@ pango_font_find_shaper (PangoFont     *font,
  * coordinate system for each rectangle has its origin at the
  * base line and horizontal origin of the character with increasing
  * coordinates extending to the right and down. The macros PANGO_ASCENT(),
- * PANGO_DESCENT(), PANGO_LBEARING(), and PANGO_RBEARING can be used to convert
+ * PANGO_DESCENT(), PANGO_LBEARING(), and PANGO_RBEARING() can be used to convert
  * from the extents rectangle to more traditional font metrics. The units
  * of the rectangles are in 1/PANGO_SCALE of a device unit.
  **/
@@ -1092,7 +1169,30 @@ pango_font_get_glyph_extents  (PangoFont      *font,
 			       PangoRectangle *ink_rect,
 			       PangoRectangle *logical_rect)
 {
-  g_return_if_fail (font != NULL);
+  if (G_UNLIKELY (!PANGO_IS_FONT (font)))
+    {
+      
+      if (!_pango_warning_history.get_glyph_extents)
+        {
+	  _pango_warning_history.get_glyph_extents = TRUE;
+	  g_warning (bad_font_warning, "pango_font_get_glyph_extents");
+	}
+      if (ink_rect)
+        {
+	  ink_rect->x = PANGO_SCALE;
+	  ink_rect->y = - (PANGO_UNKNOWN_GLYPH_HEIGHT - 1) * PANGO_SCALE;
+	  ink_rect->height = (PANGO_UNKNOWN_GLYPH_HEIGHT - 2) * PANGO_SCALE;
+	  ink_rect->width = (PANGO_UNKNOWN_GLYPH_WIDTH - 2) * PANGO_SCALE;
+        }
+      if (logical_rect)
+        {
+	  logical_rect->x = logical_rect->y = 0;
+	  logical_rect->y = - PANGO_UNKNOWN_GLYPH_HEIGHT * PANGO_SCALE;
+	  logical_rect->height = PANGO_UNKNOWN_GLYPH_HEIGHT * PANGO_SCALE;
+	  logical_rect->width = PANGO_UNKNOWN_GLYPH_WIDTH * PANGO_SCALE;
+	}
+      return;
+    }
 
   PANGO_FONT_GET_CLASS (font)->get_glyph_extents (font, glyph, ink_rect, logical_rect);
 }
@@ -1109,23 +1209,64 @@ pango_font_get_glyph_extents  (PangoFont      *font,
  * be provided to indicate that the metrics should be retrieved that
  * correspond to the script(s) used by that language.
  *
- * Returns: a #PangoMetrics object. The caller must call pango_font_metrics_unref()
+ * Return value: a #PangoFontMetrics object. The caller must call pango_font_metrics_unref()
  *   when finished using the object.
  **/
 PangoFontMetrics *
 pango_font_get_metrics (PangoFont        *font,
 			PangoLanguage    *language)
 {
+  if (G_UNLIKELY (!PANGO_IS_FONT (font)))
+    {
+      
+      if (!_pango_warning_history.get_metrics)
+        {
+	  _pango_warning_history.get_metrics = TRUE;
+	  g_warning (bad_font_warning, "pango_font_get_metrics");
+	}
+      return pango_font_metrics_new ();
+    }
+
   return PANGO_FONT_GET_CLASS (font)->get_metrics (font, language);
 }
 
+/**
+ * pango_font_get_font_map:
+ * @font: a #PangoFont
+ * 
+ * Gets the font map for which the font was created.
+ *
+ * Return value: the #PangoFontMap for the font
+ *
+ * Since: 1.10
+ **/
+PangoFontMap *
+pango_font_get_font_map (PangoFont *font)
+{
+  if (G_UNLIKELY (!PANGO_IS_FONT (font)))
+    {
+      
+      if (!_pango_warning_history.get_font_map)
+        {
+	  _pango_warning_history.get_font_map = TRUE;
+	  g_warning (bad_font_warning, "pango_font_get_font_map");
+	}
+      return NULL;
+    }
+
+  if (PANGO_FONT_GET_CLASS (font)->get_font_map)
+    return PANGO_FONT_GET_CLASS (font)->get_font_map (font);
+  else
+    return NULL;
+}
+
 GType
 pango_font_metrics_get_type (void)
 {
   static GType our_type = 0;
   
   if (our_type == 0)
-    our_type = g_boxed_type_register_static ("PangoFontMetrics",
+    our_type = g_boxed_type_register_static (I_("PangoFontMetrics"),
 					     (GBoxedCopyFunc)pango_font_metrics_ref,
 					     (GBoxedFreeFunc)pango_font_metrics_unref);
 
@@ -1146,7 +1287,7 @@ pango_font_metrics_get_type (void)
 PangoFontMetrics *
 pango_font_metrics_new (void)
 {
-  PangoFontMetrics *metrics = g_new0 (PangoFontMetrics, 1);
+  PangoFontMetrics *metrics = g_slice_new0 (PangoFontMetrics);
   metrics->ref_count = 1;
 
   return metrics;
@@ -1156,14 +1297,15 @@ pango_font_metrics_new (void)
  * pango_font_metrics_ref:
  * @metrics: a #PangoFontMetrics structure
  * 
- * Increases the reference count of a font metrics structure.
+ * Increase the reference count of a font metrics structure by one.
  * 
  * Return value: @metrics
  **/
 PangoFontMetrics *
 pango_font_metrics_ref (PangoFontMetrics *metrics)
 {
-  g_return_val_if_fail (metrics != NULL, NULL);
+  if (!metrics)
+    return NULL;
   
   metrics->ref_count++;
 
@@ -1174,20 +1316,21 @@ pango_font_metrics_ref (PangoFontMetrics *metrics)
  * pango_font_metrics_unref:
  * @metrics: a #PangoFontMetrics structure
  * 
- * Decreases the reference count of a font metrics structure. If
+ * Decrease the reference count of a font metrics structure by one. If
  * the result is zero, frees the structure and any associated
  * memory.
  **/
 void
 pango_font_metrics_unref (PangoFontMetrics *metrics)
 {
-  g_return_if_fail (metrics != NULL);
+  if (!metrics)
+    return;
   g_return_if_fail (metrics->ref_count > 0 );
   
   metrics->ref_count--;
   
   if (metrics->ref_count == 0)
-    g_free (metrics);
+    g_slice_free (PangoFontMetrics, metrics);
 }
 
 /**
@@ -1200,7 +1343,7 @@ pango_font_metrics_unref (PangoFontMetrics *metrics)
  * actual drawn ink. It is necessary to lay out the text to figure
  * where the ink will be.)
  * 
- * Return value: the ascent, in pango units. (1 point == #PANGO_SCALE pango units.)
+ * Return value: the ascent, in Pango units. (1 point == #PANGO_SCALE Pango units.)
  **/
 int
 pango_font_metrics_get_ascent (PangoFontMetrics *metrics)
@@ -1220,7 +1363,7 @@ pango_font_metrics_get_ascent (PangoFontMetrics *metrics)
  * actual drawn ink. It is necessary to lay out the text to figure
  * where the ink will be.)
  * 
- * Return value: the descent, in pango units. (1 point == #PANGO_SCALE pango units.)
+ * Return value: the descent, in Pango units. (1 point == #PANGO_SCALE Pango units.)
  **/
 int
 pango_font_metrics_get_descent (PangoFontMetrics *metrics)
@@ -1239,7 +1382,7 @@ pango_font_metrics_get_descent (PangoFontMetrics *metrics)
  * determining the initial size for a window. Actual characters in
  * text will be wider and narrower than this.
  * 
- * Return value: the character width, in pango units. (1 point == #PANGO_SCALE pango units.)
+ * Return value: the character width, in Pango units. (1 point == #PANGO_SCALE Pango units.)
  **/
 int
 pango_font_metrics_get_approximate_char_width (PangoFontMetrics *metrics)
@@ -1260,7 +1403,7 @@ pango_font_metrics_get_approximate_char_width (PangoFontMetrics *metrics)
  * is generally somewhat more accurate than the result of
  * pango_font_metrics_get_approximate_digit_width().
  * 
- * Return value: the digit width, in pango units. (1 point == #PANGO_SCALE pango units.)
+ * Return value: the digit width, in Pango units. (1 point == #PANGO_SCALE Pango units.)
  **/
 int
 pango_font_metrics_get_approximate_digit_width (PangoFontMetrics *metrics)
@@ -1270,36 +1413,98 @@ pango_font_metrics_get_approximate_digit_width (PangoFontMetrics *metrics)
   return metrics->approximate_digit_width;
 }
 
+/**
+ * pango_font_metrics_get_underline_position:
+ * @metrics: a #PangoFontMetrics structure
+ * 
+ * Gets the suggested position to draw the underline.
+ * The value returned is the distance <emphasis>above</emphasis> the
+ * baseline of the top of the underline. Since most fonts have
+ * underline positions beneath the baseline, this value is typically
+ * negative.
+ * 
+ * Return value: the suggested underline position, in Pango units.
+ *
+ * Since: 1.6
+ **/
+int
+pango_font_metrics_get_underline_position (PangoFontMetrics *metrics)
+{
+  g_return_val_if_fail (metrics != NULL, 0);
+
+  return metrics->underline_position;
+}
+
+/**
+ * pango_font_metrics_get_underline_thickness:
+ * @metrics: a #PangoFontMetrics structure
+ * 
+ * Gets the suggested thickness to draw for the underline.
+ * 
+ * Return value: the suggested underline thickness, in Pango units.
+ *
+ * Since: 1.6
+ **/
+int
+pango_font_metrics_get_underline_thickness (PangoFontMetrics *metrics)
+{
+  g_return_val_if_fail (metrics != NULL, 0);
+
+  return metrics->underline_thickness;
+}
+
+/**
+ * pango_font_metrics_get_strikethrough_position:
+ * @metrics: a #PangoFontMetrics structure
+ * 
+ * Gets the suggested position to draw the strikethrough.
+ * The value returned is the distance <emphasis>above</emphasis> the
+ * baseline of the top of the strikethrough.
+ * 
+ * Return value: the suggested strikethrough position, in Pango units.
+ *
+ * Since: 1.6
+ **/
+int
+pango_font_metrics_get_strikethrough_position (PangoFontMetrics *metrics)
+{
+  g_return_val_if_fail (metrics != NULL, 0);
+
+  return metrics->strikethrough_position;
+}
+
+/**
+ * pango_font_metrics_get_strikethrough_thickness:
+ * @metrics: a #PangoFontMetrics structure
+ * 
+ * Gets the suggested thickness to draw for the strikethrough.
+ * 
+ * Return value: the suggested strikethrough thickness, in Pango units.
+ *
+ * Since: 1.6
+ **/
+int
+pango_font_metrics_get_strikethrough_thickness (PangoFontMetrics *metrics)
+{
+  g_return_val_if_fail (metrics != NULL, 0);
+
+  return metrics->strikethrough_thickness;
+}
+
 /*
  * PangoFontFamily
  */
 
-GType
-pango_font_family_get_type (void)
-{
-  static GType object_type = 0;
+G_DEFINE_TYPE (PangoFontFamily, pango_font_family, G_TYPE_OBJECT)
 
-  if (!object_type)
-    {
-      static const GTypeInfo object_info =
-      {
-        sizeof (PangoFontFamilyClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        NULL,           /* class_init */
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoFontFamily),
-        0,              /* n_preallocs */
-	NULL            /* init */
-      };
-      
-      object_type = g_type_register_static (G_TYPE_OBJECT,
-                                            "PangoFontFamily",
-                                            &object_info, 0);
-    }
-  
-  return object_type;
+static void
+pango_font_family_class_init (PangoFontFamilyClass *class)
+{
+}
+
+static void
+pango_font_family_init (PangoFontFamily *family)
+{
 }
 
 /**
@@ -1323,7 +1528,7 @@ pango_font_family_get_name (PangoFontFamily  *family)
 
 /**
  * pango_font_family_list_faces:
- * @family: a #PangFontFamily
+ * @family: a #PangoFontFamily
  * @faces: location to store an array of pointers to #PangoFontFace
  *   objects, or %NULL. This array should be freed with g_free()
  *   when it is no longer needed.
@@ -1343,36 +1548,52 @@ pango_font_family_list_faces (PangoFontFamily  *family,
   PANGO_FONT_FAMILY_GET_CLASS (family)->list_faces (family, faces, n_faces);
 }
 
+/**
+ * pango_font_family_is_monospace:
+ * @family: a #PangoFontFamily
+ *
+ * A monospace font is a font designed for text display where the the
+ * characters form a regular grid. For Western languages this would
+ * mean that the advance width of all characters are the same, but
+ * this categorization also includes Asian fonts which include
+ * double-width characters: characters that occupy two grid cells.
+ * g_unichar_iswide() returns a result that indicates whether a
+ * character is typically double-width in a monospace font.
+ * 
+ * The best way to find out the grid-cell size is to call
+ * pango_font_metrics_get_approximate_digit_width(), since the results
+ * of pango_font_metrics_get_approximate_char_width() may be affected
+ * by double-width characters.  
+ * 
+ * Return value: %TRUE if the family is monospace.
+ *
+ * Since: 1.4
+ **/
+gboolean 
+pango_font_family_is_monospace (PangoFontFamily  *family)
+{
+  g_return_val_if_fail (PANGO_IS_FONT_FAMILY (family), FALSE);
+
+  if (PANGO_FONT_FAMILY_GET_CLASS (family)->is_monospace)
+    return PANGO_FONT_FAMILY_GET_CLASS (family)->is_monospace (family);
+  else
+    return FALSE;
+}
+
 /*
  * PangoFontFace
  */
 
-GType
-pango_font_face_get_type (void)
-{
-  static GType object_type = 0;
+G_DEFINE_TYPE (PangoFontFace, pango_font_face, G_TYPE_OBJECT)
 
-  if (!object_type)
-    {
-      static const GTypeInfo object_info =
-      {
-        sizeof (PangoFontFaceClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        NULL,           /* class_init */
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoFont),
-        0,              /* n_preallocs */
-	NULL            /* init */
-      };
-      
-      object_type = g_type_register_static (G_TYPE_OBJECT,
-                                            "PangoFontFace",
-                                            &object_info, 0);
-    }
-  
-  return object_type;
+static void
+pango_font_face_class_init (PangoFontFaceClass *class)
+{
+}
+
+static void
+pango_font_face_init (PangoFontFace *face)
+{
 }
 
 /**
@@ -1414,3 +1635,37 @@ pango_font_face_get_face_name (PangoFontFace *face)
 
   return PANGO_FONT_FACE_GET_CLASS (face)->get_face_name (face);
 }
+
+/**
+ * pango_font_face_list_sizes:
+ * @face: a #PangoFontFace.
+ * @sizes: location to store a pointer to an array of int. This array
+ *         should be freed with g_free().
+ * @n_sizes: location to store the number of elements in @sizes
+ * 
+ * List the available sizes for a font. This is only applicable to bitmap
+ * fonts. For scalable fonts, stores %NULL at the location pointed to by
+ * @sizes and 0 at the location pointed to by @n_sizes. The sizes returned
+ * are in Pango units and are sorted in ascending order.
+ *
+ * Since: 1.4
+ **/
+void
+pango_font_face_list_sizes (PangoFontFace  *face, 
+                            int           **sizes, 
+                            int            *n_sizes)
+{
+  g_return_if_fail (PANGO_IS_FONT_FACE (face));
+  g_return_if_fail (sizes == NULL || n_sizes != NULL);
+
+  if (n_sizes == NULL)
+    return;
+
+  if (PANGO_FONT_FACE_GET_CLASS (face)->list_sizes != NULL)
+    PANGO_FONT_FACE_GET_CLASS (face)->list_sizes (face, sizes, n_sizes);
+  else
+    {
+      *sizes = NULL;
+      *n_sizes = 0;
+    }
+}
